#!/usr/bin/python3

from collections import defaultdict

class Parser:
    def __init__(self, lexrulefile, gramrulefile):
        self.lexrules = self.read_lexicon(lexrulefile)
        self.gramrules, self.startsymbol = self.read_grammar(gramrulefile)
        
    @staticmethod
    def read_lexicon(filename):
        lexrules = defaultdict(list)
        with open(filename) as file:
            for line in file:
                prob, tag, word = line.split()
                lexrules[word].append((tag, float(prob)))
        return dict(lexrules)
    
    @staticmethod
    def read_grammar(filename):
        gramrules = []
        with open(filename) as file:
            for line in file:
                prob, A, B, C = line.split()
                gramrules.append(((A,B,C), float(prob)))
        return gramrules, gramrules[0][0][0]
    
    def parse(self, words):
        n = len(words)
        vitprob, bestparse = {}, {}
        # lexicon lookup
        for i, word in enumerate(words):
            for tag, p in self.lexrules.get(word, []):
                vitprob[i, i+1, tag] = p
                bestparse[i, i+1, tag] = f"({tag} {word})"
        # parsing
        for k in range(2, n+1):
            for i in range(k-2, -1, -1):
                for j in range(i+1, k):
                    for (A,B,C), p in self.gramrules:
                        p1 = vitprob.get((i, j, B), 0.0)
                        p2 = vitprob.get((j, k, C), 0.0)
                        p = p * p1 * p2
                        if vitprob.get((i, k, A), 0.0) < p:
                            vitprob[i, k, A] = p
                            t1 = bestparse[i, j, B]
                            t2 = bestparse[j, k, C]
                            bestparse[i, k, A] = f"({A} {t1} {t2})"
        parse = bestparse.get((0, n, self.startsymbol),
                              "Die Eingabe ist ungrammatisch")
        return parse

parser = Parser("lexrules.txt", "gramrules.txt")
parse = parser.parse("die Erde ist ein Planet".split())
print(parse)
